module cloud_majcomp

  use vtype_module

  implicit none

  private

  public :: CloudMajComp

contains

  !--------------------------------------------------------------------------------------

  subroutine CloudMajComp(                         &
    & FlagSedimentation, &
    & FlagSedimentationInfty, &
    & Grav, Cp, MajCompLatentHeat, CloudNumRatio, TimeStep, &
    & kmax, &
    & r_Press, z_Press, r_Height, &
    & z_CldDen, z_CldMMR, z_CldRad, r_Temp &
    & )
    !
    ! CO2 相変化
    !
    ! CO2 phase change
    !

    ! モジュール引用 ; USE statements
    !
    use saturation_K1991, only : SatK1991CalcTempSat


    ! 宣言文 ; Declaration statements
    !
    implicit none

    logical , intent(in   ) :: FlagSedimentation
    logical , intent(in   ) :: FlagSedimentationInfty
    real(dp), intent(in   ) :: Grav
    real(dp), intent(in   ) :: Cp
    real(dp), intent(in   ) :: MajCompLatentHeat
    real(dp), intent(in   ) :: CloudNumRatio
    real(dp), intent(in   ) :: TimeStep
    integer , intent(in   ) :: kmax
    real(dp), intent(in   ) :: r_Press     (0:kmax)
    real(dp), intent(in   ) :: z_Press     (1:kmax)
    real(dp), intent(in   ) :: r_Height    (0:kmax)
    real(dp), intent(in   ) :: z_CldDen    (1:kmax)
    real(dp), intent(inout) :: z_CldMMR    (1:kmax)
    real(dp), intent(inout) :: z_CldRad    (1:kmax)
    real(dp), intent(inout) :: r_Temp      (0:kmax)


    ! 作業変数
    ! Work variables
    !
    real(DP):: z_DelAtmMass  (1:kmax)
    real(DP):: z_CldIceMass  (1:kmax)

    real(DP):: z_Temp            (1:kmax)

    real(DP):: z_TempCond        (1:kmax)

    real(DP):: r_GravSedMassFlux (0:kmax)
    real(DP):: r_DPPress         (0:kmax)


    integer:: k              ! 鉛直方向に回る DO ループ用作業変数
                             ! Work variables for DO loop in vertical direction


    ! 実行文 ; Executable statement
    !


    do k = 1, kmax
      z_DelAtmMass(k) = ( r_Press(k-1) - r_Press(k) ) / Grav
    end do


    call CloudMajCompPhaseChange( &
      & Cp, MajCompLatentHeat, CloudNumRatio, & ! (in   )
      & kmax, r_Press, z_DelAtmMass,          & ! (in   )
      & z_CldMMR, z_CldRad, r_Temp            & ! (inout)
      & )

    do k = 1, kmax
      z_TempCond(k) = SatK1991CalcTempSat( z_Press(k) )
    end do

    if ( FlagSedimentation ) then


      ! 重力沈降過程
      ! Gravitational sedimentation process
      !
      z_Temp = r_Temp(1:kmax)
      z_CldIceMass = z_DelAtmMass * z_CldMMR
      call CloudMajorCompCalcSedMassFlux1D( &
        & FlagSedimentationInfty,          & ! (in   )
        & TimeStep,                        & ! (in   )
        & Grav, Cp,                        & ! (in   )
        & MajCompLatentHeat,               & ! (in   )
        & kmax,                            & ! (in   )
        & r_Press, r_Height,               & ! (in   )
        & z_CldDen, z_CldRad,              & ! (in   )
        & z_TempCond,                      & ! (in   )
        & z_Temp,                          & ! (inout)
        & z_CldIceMass,                    & ! (inout)
        & r_GravSedMassFlux,               & ! (out  )
        & r_DPPress                        & ! (out  )
        & )
      r_Temp(1:kmax) = z_Temp
      z_CldMMR = z_CldIceMass / z_DelAtmMass

      if ( r_GravSedMassFlux(0) < 0.0d0 ) then
        write( 6, * ) 'Atmosphere will collapse in CloudMajComp'
        write( 6, * ) '  Sedimentation flux is not zero at the surface : ', &
          & r_GravSedMassFlux(0)
        stop
      end if

      call CloudMajCompPhaseChange( &
        & Cp, MajCompLatentHeat, CloudNumRatio, & ! (in   )
        & kmax, r_Press, z_DelAtmMass,          & ! (in   )
        & z_CldMMR, z_CldRad, r_Temp            & ! (inout)
        & )

      ! T0 = ( T2 - T1 ) / ( 2 - 1 ) * ( 0 - 1 ) + T1
      !    = - ( T2 - T1 ) + T1
      !    = 2 * T1 - T2
      r_Temp(0) = 2.0d0 * r_Temp(1) - r_Temp(2)
    end if


!!$    xyr_DPressDt         = xyr_GravSedMassFlux * Grav
!!$    xy_DPsDt             = xyr_DPressDt(:,:,0)
!!$    xy_DSurfMajCompIceDt = - xy_DPsDt / Grav


  end subroutine CloudMajComp

  !------------------------------------------------------------------------------------

  subroutine CloudMajorCompCalcSedMassFlux1D( &
    & FlagSedimentationInfty,          & ! (in   )
    & TimeStep,                        & ! (in   )
    & Grav, CpDry,                     & ! (in   )
    & LatentHeatMajCompSubl,           & ! (in   )
    & kmax,                            & ! (in   )
    & r_Press, r_Height,               & ! (in   )
    & z_PartDen, z_PartRad,            & ! (in   )
    & z_TempCond,                      & ! (in   )
    & z_Temp,                          & ! (inout)
    & z_MajCompIceMass,                & ! (inout)
    & r_GravSedMassFlux,               & ! (out  )
    & r_DPPress                        & ! (out  )
    & )

    ! 物理・数学定数設定
    ! Physical and mathematical constants settings
    !
    use constants0, only: &
      & PI                    ! $ \pi $ .
                              ! 円周率.  Circular constant

    logical     , intent(in   ) :: FlagSedimentationInfty
    real(DP)    , intent(in   ) :: TimeStep
    real(DP)    , intent(in   ) :: Grav
    real(DP)    , intent(in   ) :: CpDry
    real(DP)    , intent(in   ) :: LatentHeatMajCompSubl
    integer     , intent(in   ) :: kmax
    real(DP)    , intent(in   ) :: r_Press   (0:kmax)
    real(DP)    , intent(in   ) :: r_Height  (0:kmax)
    real(DP)    , intent(in   ) :: z_PartDen (1:kmax)
    real(DP)    , intent(in   ) :: z_PartRad (1:kmax)
    real(DP)    , intent(in   ) :: z_TempCond(1:kmax)
    real(DP)    , intent(inout) :: z_Temp    (1:kmax)
    real(DP)    , intent(inout) :: z_MajCompIceMass (1:kmax)
    real(DP)    , intent(out  ) :: r_GravSedMassFlux(0:kmax)
    real(DP)    , intent(out  ) :: r_DPPress        (0:kmax)


    !
    ! local variables
    !
    real(dp) :: r_PartDen     (0:kmax)
    real(dp) :: r_PartRad     (0:kmax)

    real(DP) :: z_DelAtmMass    (1:kmax)
    real(DP) :: z_DelCompMass   (1:kmax)
    real(DP) :: FallingCompMass
    real(DP) :: FallingCompMassSubl
!!$    real(DP) :: DelCompMassTentative
    real(DP) :: z_CompMassSink  (1:kmax)

    real(DP) :: r_SedVel        (0:kmax)
    real(DP) :: r_GravSedMass   (0:kmax)

    real(DP) :: r_DPressDt      (0:kmax)

    real(DP) :: APHeight
    real(DP) :: DPHeight

    real(DP) :: MassOfPartialLayer
    real(DP) :: MassConsumed

    real(DP) :: Press1
    real(DP) :: Press2
    real(DP) :: Press

    real(DP) :: DelTempSubl1
    real(DP) :: DelTempSubl2
    real(DP) :: DelTempSubl

    integer  :: k
    integer  :: kk
    integer  :: kkk


    ! 実行文 ; Executable statement
    !


    ! Calculation of mass in each layer and layer thickness in unit of meter
    !   Layer thickness is calculated by using mass of a layer.
    do k = 1, kmax
      z_DelAtmMass(k) = ( r_Press(k-1) - r_Press(k) ) / Grav
    end do

    z_DelCompMass = z_MajCompIceMass

    !
    ! calculation of sedimentation terminal velocity
    !
    if ( FlagSedimentationInfty ) then
      r_SedVel = -1.0d100
    else
      r_PartDen(0:kmax-1) = z_PartDen(1:kmax)
      r_PartDen(kmax)     = z_PartDen(kmax)
      r_PartRad(0:kmax-1) = z_PartRad(1:kmax)
      r_PartRad(kmax)     = z_PartRad(kmax)
      call CloudMajCompCalcSedVel( &
        & Grav, &
        & kmax+1, r_Press, r_PartDen, r_PartRad, &
        & r_SedVel &
        & )
    end if

    ! arbitrary upper (lower) limit
!!$        r_SedVel = sign( 1.0_DP, r_SedVel ) * max( abs(r_SedVel), 1.0e-20_DP )


    ! Evaluation of gravitational sedimentation mass

    !   Initialization
    r_GravSedMass   = 0.0_DP
    z_CompMassSink  = 0.0_DP

    ! kk : layer index
    do kk = 1, kmax

      ! Initialization
      FallingCompMass = z_DelCompMass(kk)

      ! k : level index
      loop_level_k : do k = kk-1, 0, -1

        ! mass is sublimed in (kkk)th = (k+1)th layer during sedimentation
        !   kkk : layer index
        kkk = k+1
        !   sublimation
        !     temperature decrease when only part of ice sublime
        DelTempSubl1 = max( z_Temp(kkk) - z_TempCond(kkk), 0.0_DP )
        !     temperature decrease when all of ice sublime
        DelTempSubl2 = &
          &   LatentHeatMajCompSubl * FallingCompMass &
          &   / ( CpDry * z_DelAtmMass(kkk) )
        DelTempSubl = min( DelTempSubl1, DelTempSubl2 )
        z_Temp(kkk) = z_Temp(kkk) - DelTempSubl
        FallingCompMassSubl =                         &
          &   CpDry * z_DelAtmMass(kkk) * DelTempSubl &
          &   / LatentHeatMajCompSubl
        FallingCompMass     = FallingCompMass     - FallingCompMassSubl
        z_CompMassSink(kkk) = z_CompMassSink(kkk) + FallingCompMassSubl


        ! check whether (kk-1)th level crosses kth level
        !   note : (kk-1)th level is a bottom of (kk)th layer
        !   note : APHeight : height of arrival point
        APHeight = r_Height(kk-1) + r_SedVel(kk-1) * TimeStep
        if ( r_Height(k) <= APHeight ) then
          ! (kk-1)th level does not cross kth level
          FallingCompMass = 0.0_DP
          exit loop_level_k
        else
          ! (kk-1)th level crosses kth level

          ! check whether (kk)th level crosses kth level
          !   note : (kk)th level is a top of (kk)th layer
          !   Here, sedimentation velocity in (kk)th layer is assumed to be
          !   constant with a velocity at (kk-1)th level (bottom level of a
          !   layer).
          !   So, sedimentation velocity at (kk-1)th level is used.
          !   note : r_SedVel is negative
          !   note : APHeight : height of arrival point
          APHeight = r_Height(kk) + r_SedVel(kk-1) * TimeStep
          if ( APHeight <= r_Height(k) ) then
            ! All mass in (kk)th layer cross kth level

            ! do nothing
          else
            ! Part of mass in (kk)th layer crosses kth level
            ! note : r_SedVel is negative

            ! note : DPHeight : height of departure point, i.e., height which
            !                   reaches kth level at end of a time step
            DPHeight = r_Height(k) - r_SedVel(kk-1) * TimeStep

            ! evaluate pressure at the departure point
            !   Press1 : pressure at bottom of (kk)th layer
            !   Press2 : pressure at top    of (kk)th layer
            Press1 = r_Press(kk-1)
            if ( r_Press(kk) == 0.0_DP ) then
              Press2 = r_Press(kk-1) * 1.0e-1_DP
            else
              Press2 = r_Press(kk  )
            end if
            !   Press  : pressure at the departure point
            Press = &
              &   log( Press2 / Press1 )              &
              & / ( r_Height  (kk) - r_Height(kk-1) ) &
              & * ( DPHeight       - r_Height(kk-1) ) &
              & + log( Press1 )
            Press = exp( Press )
!!$            DelCompMass =              &
!!$              &   ( Press1 - Press  )  &
!!$              & / ( Press1 - Press2 )  &
!!$              & * z_DelCompMass(kk)
            ! MassOfPartialLayer : Mass between Press1 and Press, i.e., mass
            !                    : which crosses kth level
            MassOfPartialLayer =       &
              &   ( Press1 - Press  )  &
              & / ( Press1 - Press2 )  &
              & * z_DelCompMass(kk)
            ! MassConsumed : Mass of (kk)th layer which is consumed (sublimed)
            ! in layers above
            MassConsumed = z_DelCompMass(kk) - FallingCompMass
            ! Concept:
            !   MassOfPartialLayer is mass of bottom part of (kk)th layer.
            !   Mass of bottom part of layer will be consumed (sublimed) in
            !   layers above, first (preferentially).
            FallingCompMass = max( MassOfPartialLayer - MassConsumed, 0.0_DP )
          end if

        end if

        ! Gravitational sedimentation mass is increased by contribution of
        ! (kk)th layer (remnant of (kk)th layer mass).
        r_GravSedMass(k) = r_GravSedMass(k) + FallingCompMass

      end do loop_level_k

    end do


!!$    ! Modification of sedimentation mass to avoid negative mass
!!$    do k = kmax, 1, -1
!!$      ! note: negative sign on r_GravSedMass means downward transport
!!$      DelCompMassTentative = z_DelCompMass(k) &
!!$        & - ( ( -r_GravSedMass(k) ) - ( -r_GravSedMass(k-1) ) ) &
!!$        & - z_CompMassSink(k)
!!$      if ( DelCompMassTentative < 0.0_DP ) then
!!$        r_GravSedMass(k-1) = z_DelCompMass(k) + r_GravSedMass(k) - z_CompMassSink(k)
!!$        z_DelCompMass(k) = 0.0_DP
!!$      else
!!$        z_DelCompMass(k) = DelCompMassTentative
!!$      end if
!!$    end do
    do k = 1, kmax
      ! note: negative sign on r_GravSedMass means downward transport
      z_DelCompMass(k) = z_DelCompMass(k)                       &
        & - ( ( -r_GravSedMass(k) ) - ( -r_GravSedMass(k-1) ) ) &
        & - z_CompMassSink(k)
    end do

    ! note: negative sign on r_GravSedMass means downward transport
    r_GravSedMassFlux = - r_GravSedMass / TimeStep

    z_MajCompIceMass = z_DelCompMass

    r_DPressDt = - r_GravSedMassFlux * Grav
    r_DPPress = r_Press + r_DPressDt * TimeStep


  end subroutine CloudMajorCompCalcSedMassFlux1D

  !----------------------------------------------------------------------------
  ! NOT USED NOW
  subroutine CloudMajCompExp( &
    & FlagSedimentation, &
    & Grav, Cp, MajCompLatentHeat, CloudNumRatio, TimeStep, &
    & kmax, r_Press, r_Height, z_CldDen, &
    & z_CldMMR, z_CldRad, r_Temp &
    & )

    logical , intent(in   ) :: FlagSedimentation
    real(dp), intent(in   ) :: Grav
    real(dp), intent(in   ) :: Cp
    real(dp), intent(in   ) :: MajCompLatentHeat
    real(dp), intent(in   ) :: CloudNumRatio
    real(dp), intent(in   ) :: TimeStep
    integer , intent(in   ) :: kmax
    real(dp), intent(in   ) :: r_Press     (0:kmax)
    real(dp), intent(in   ) :: r_Height    (0:kmax)
    real(dp), intent(in   ) :: z_CldDen    (1:kmax)
    real(dp), intent(inout) :: z_CldMMR    (1:kmax)
    real(dp), intent(inout) :: z_CldRad    (1:kmax)
    real(dp), intent(inout) :: r_Temp      (0:kmax)


    ! Local variables
    !
    real(dp) :: r_CldDen     (0:kmax)
    real(dp) :: r_CldRad     (0:kmax)
    real(dp) :: r_SedVel     (0:kmax)
    real(dp) :: z_DelAtmMass (1:kmax)
    real(dp) :: z_DelCldMass (1:kmax)
    real(dp) :: z_DelCldMassA(1:kmax)

    real(dp) :: CourantNum
    real(dp) :: TimeStepSubLoop
    integer  :: NLoop
    integer  :: iLoop

    integer  :: k


    do k = 1, kmax
      z_DelAtmMass(k) = ( r_Press(k-1) - r_Press(k) ) / Grav
    end do


    call CloudMajCompPhaseChange( &
      & Cp, MajCompLatentHeat, CloudNumRatio, &
      & kmax, r_Press, z_DelAtmMass, &
      & z_CldMMR, z_CldRad, r_Temp &
      & )


    if ( FlagSedimentation ) then
      r_CldDen(0:kmax-1) = z_CldDen(1:kmax)
      r_CldDen(kmax)     = z_CldDen(kmax)
      r_CldRad(0:kmax-1) = z_CldRad(1:kmax)
      r_CldRad(kmax)     = z_CldRad(kmax)
      call CloudMajCompCalcSedVel( &
        & Grav, &
        & kmax+1, r_Press, r_CldDen, r_CldRad, &
        & r_SedVel &
        & )

      ! check
      do k = 0, kmax-1
        CourantNum =   ( abs( r_SedVel(k) ) * TimeStep ) &
          &          / ( r_Height(k+1) - r_Height(k) )
        if ( CourantNum > 1.0d0 ) then
          write( 6, * ) 'warning: CFL', k, CourantNum
        end if
      end do

      ! sub-loop
      NLoop = 1
      TimeStepSubLoop = TimeStep / dble( NLoop )
      do iLoop = 1, NLoop

        r_CldDen(0:kmax-1) = z_CldDen(1:kmax)
        r_CldDen(kmax)     = z_CldDen(kmax)
        r_CldRad(0:kmax-1) = z_CldRad(1:kmax)
        r_CldRad(kmax)     = z_CldRad(kmax)
        call CloudMajCompCalcSedVel( &
          & Grav, &
          & kmax+1, r_Press, r_CldDen, r_CldRad, &
          & r_SedVel &
          & )

        z_DelCldMass = z_DelAtmMass * z_CldMMR
        call CloudMajCompSedimentation( &
          & TimeStep, &
          & kmax, r_Press, r_Height, r_SedVel, z_DelCldMass, &
          & z_DelCldMassA &
          & )
        z_DelCldMass = z_DelCldMassA
        z_CldMMR = z_DelCldMass / z_DelAtmMass

        call CloudMajCompPhaseChange( &
          & Cp, MajCompLatentHeat, CloudNumRatio, &
          & kmax, r_Press, z_DelAtmMass, &
          & z_CldMMR, z_CldRad, r_Temp &
          & )
      end do
    end if


  end subroutine CloudMajCompExp

  !----------------------------------------------------------------------------

  subroutine CloudMajCompPhaseChange( &
    & Cp, MajCompLatentHeat, CloudNumRatio, & ! (in   )
    & kmax, r_Press, z_DelAtmMass,          & ! (in   )
    & z_CldMMR, z_CldRad, r_Temp            & ! (inout)
    & )

    use constants0, only : PI

    use saturation_K1991, only : SatK1991CalcTempSat

    real(dp), intent(in   ) :: Cp
    real(dp), intent(in   ) :: MajCompLatentHeat
    real(dp), intent(in   ) :: CloudNumRatio
    integer , intent(in   ) :: kmax
    real(dp), intent(in   ) :: r_Press     (0:kmax)
    real(dp), intent(in   ) :: z_DelAtmMass(1:kmax)
    real(dp), intent(inout) :: z_CldMMR    (1:kmax)
    real(dp), intent(inout) :: z_CldRad    (1:kmax)
    real(dp), intent(inout) :: r_Temp      (0:kmax)


    ! Local variables
    !
    real(dp) :: z_DelCldMass(1:kmax)
    real(dp) :: DelCldMass
    real(dp) :: TempSat

    integer  :: k


    z_DelCldMass = z_DelAtmMass * z_CldMMR


    do k = kmax, 1, -1
      DelCldMass = z_DelCldMass(k)
      ! Sublimation
      r_Temp(k) = r_Temp(k) &
        & - DelCldMass * MajCompLatentHeat / ( Cp * z_DelAtmMass(k) )
      DelCldMass = 0.0d0
      !
      TempSat = SatK1991CalcTempSat( r_Press(k) )
      if ( r_Temp(k) < TempSat ) then ! Condensation
        DelCldMass = ( TempSat - r_Temp(k) ) * Cp * z_DelAtmMass(k) / MajCompLatentHeat
        r_Temp(k) = TempSat
        write( 6, * ) 'Atmosphere condenses at k = ', k, ' : ', r_Press(k), r_Temp(k), TempSat
      end if

      z_CldMMR(k) = DelCldMass / z_DelAtmMass(k)
      ! M_ptcl = N_ptcl * 4/3*PI*R**3 * Rho
      ! R**3 = M_ptcl / ( N_ptcl * 4/3*PI * Rho )
      ! R = ( M_ptcl / ( N_ptcl * 4/3*PI * Rho ) )**(1/3)
      ! R = ( M_ptcl / ( M_atm * Ratio * 4/3*PI * Rho ) )**(1/3)
      ! 1.62e3 : CO2 density at 150 K (Mangan et al., 2017)
      z_CldRad(k) = &
        & ( &
        &     DelCldMass &
        &   / ( z_DelAtmMass(k) * CloudNumRatio * 4.0d0/3.0d0 * PI * 1.62e3 ) &
        & )**(1.0d0/3.0d0) &
        & + 1.0d-30  ! arbitrary lower limit
    end do


  end subroutine CloudMajCompPhaseChange

  !----------------------------------------------------------------------------

  subroutine CloudMajCompCalcSedVel( &
    & Grav, &
    & kmax, z_Press, z_PartDen, z_PartRad, &
    & z_SedVel &
    & )


    real(dp), intent(in ) :: Grav
    integer , intent(in ) :: kmax
    real(dp), intent(in ) :: z_Press   (1:kmax)
    real(dp), intent(in ) :: z_PartDen (1:kmax)
    real(dp), intent(in ) :: z_PartRad (1:kmax)
    real(dp), intent(out) :: z_SedVel  (1:kmax)


    ! Local variables
    !
    real(dp) :: z_PartDia     (1:kmax)
    real(dp) :: z_MeanFreePath(1:kmax)
    real(dp) :: MolVisCoef
    real(dp) :: MeanFreePathRef
    real(dp) :: PressLambdaRef


    z_PartDia = 2.0d0 * z_PartRad

    !
    ! The values below are obtained from Conrath (1975).
    ! Particle radius of 1e-6 m is assumed.
    !
    MolVisCoef      = 1.5e-4_DP * 1.0e-3_DP * 1.0e2_DP
    MeanFreePathRef = 2.2e-4_DP * 1.0e-2_DP
    PressLambdaRef  = 25.0e2_DP

    z_MeanFreePath = MeanFreePathRef * ( PressLambdaRef / max( z_Press, 1.0d-10 ) )

    z_SedVel = a_SedVel( &
      & Grav, kmax, z_PartDen, z_PartDia, MolVisCoef, z_MeanFreePath &
      & )

  end subroutine CloudMajCompCalcSedVel

  !----------------------------------------------------------------------------

  subroutine CloudMajCompSedimentation( &
    & TimeStep, &
    & kmax, r_Press, r_Height, r_SedVel, z_DelMass, &
    & z_DelMassA &
    & )


    real(dp), intent(in ) :: TimeStep
    integer , intent(in ) :: kmax
    real(dp), intent(in ) :: r_Press   (0:kmax)
    real(dp), intent(in ) :: r_Height  (0:kmax)
    real(dp), intent(in ) :: r_SedVel  (0:kmax)
    real(dp), intent(in ) :: z_DelMass (1:kmax)
    real(dp), intent(out) :: z_DelMassA(1:kmax)


    ! Local variables
    !
    real(dp) :: SedVel

    real(dp) :: r_APHeight    (0:kmax)
    real(dp) :: r_DPHeight    (0:kmax)
    integer  :: r_DPIndex     (0:kmax)
    real(dp) :: r_DPPress     (0:kmax)
    real(dp) :: PressBot
    real(dp) :: PressTop
    real(dp) :: HeightBot
    real(dp) :: HeightTop
    real(dp) :: DPHeight
    real(dp) :: ScaleHeight

    integer  :: k
    integer  :: kk


    ! Assumption:
    !   * Sedimentation velocity has linear subgrid distribution.
    !   * Sedimentation velocity is evaluated at center of arrival point
    !     and departure point.
    !
    ! w = ( w_{k} - w_{k-1} ) / ( z_{k} - z_{k-1} ) * ( z_{mp} - z_{k-1} ) + w_{k-1}
    !   = ( w_{k} - w_{k-1} ) / ( z_{k} - z_{k-1} ) * ( ( z_{ap} + z_{dp} ) / 2 - z_{k-1} ) + w_{k-1}
    !   = ( w_{k} - w_{k-1} ) / ( z_{k} - z_{k-1} ) * ( z_{ap} / 2 - z_{k-1} ) + w_{k-1}
    !   + ( w_{k} - w_{k-1} ) / ( z_{k} - z_{k-1} ) * z_{dp} / 2
    !
    ! z_{dp} = z_{ap} - w * dt
    !
    ! w = ( w_{k} - w_{k-1} ) / ( z_{k} - z_{k-1} ) * ( z_{ap} / 2 - z_{k-1} ) + w_{k-1}
    !   + ( w_{k} - w_{k-1} ) / ( z_{k} - z_{k-1} ) * ( z_{ap} - w * dt ) / 2
    !
    ! ( 1 + ( w_{k} - w_{k-1} ) / ( z_{k} - z_{k-1} ) * dt / 2 ) * w
    !   = ( w_{k} - w_{k-1} ) / ( z_{k} - z_{k-1} ) * ( z_{ap} / 2 - z_{k-1} ) + w_{k-1}
    !   + ( w_{k} - w_{k-1} ) / ( z_{k} - z_{k-1} ) * z_{ap} / 2
    !   = ( w_{k} - w_{k-1} ) / ( z_{k} - z_{k-1} ) * ( z_{ap} - z_{k-1} ) + w_{k-1}
    !   = ( w_{k} - w_{k-1} ) / ( z_{k} - z_{k-1} ) * ( z_{ap} - z_{k-1} )
    !             + w_{k-1}   / ( z_{k} - z_{k-1} ) * ( z_{k}  - z_{k-1} )
    !   = ( ( w_{k} - w_{k-1} ) * ( z_{ap} - z_{k-1} )
    !               + w_{k-1}   * ( z_{k}  - z_{k-1} ) )
    !     / ( z_{k} - z_{k-1} )
    !   = ( w_{k} * ( z_{ap} - z_{k-1} ) - w_{k-1} * ( z_{ap} - z_{k-1} )
    !                                    + w_{k-1} * ( z_{k}  - z_{k-1} ) )
    !     / ( z_{k} - z_{k-1} )
    !   = ( w_{k} * ( z_{ap} - z_{k-1} ) - w_{k-1} * z_{ap}
    !                                    + w_{k-1} * z_{k}  )
    !     / ( z_{k} - z_{k-1} )
    !   = ( w_{k} * ( z_{ap} - z_{k-1} ) - w_{k-1} * ( z_{ap} - z_{k} ) )
    !     / ( z_{k} - z_{k-1} )
    !   = ( w_{k} * ( z_{ap} - z_{k-1} ) + w_{k-1} * ( z_{k} - z_{ap} ) )
    !     / ( z_{k} - z_{k-1} )


    r_APHeight = r_Height
    do k = 0, kmax-1
      ! MEMO: r_SedVel is negative, i.e., downward.
      !
      ! Simple method
!!$      SedVel = r_SedVel(k)
      !
      ! Linear method
      !     ( 1 + ( w_{k} - w_{k-1} ) / ( z_{k} - z_{k-1} ) * dt / 2 ) * w
      !   = ( w_{k} * ( z_{ap} - z_{k-1} ) + w_{k-1} * ( z_{k} - z_{ap} ) )
      !     / ( z_{k} - z_{k-1} )
      SedVel = &
        &   (   r_SedVel(k+1) * ( r_APHeight(k  ) - r_Height  (k  ) )   &
        &     + r_SedVel(k  ) * ( r_Height  (k+1) - r_APHeight(k  ) ) ) &
        & / ( r_Height(k+1) - r_Height(k) )                             &
        & / ( 1.0d0 +   ( r_SedVel(k+1) - r_SedVel(k) )                 &
        &             / ( r_Height(k+1) - r_Height(k) ) * TimeStep / 2.0d0 )
      !
      r_DPHeight(k) = r_APHeight(k) - SedVel * TimeStep
      r_DPHeight(k) = min( r_DPHeight(k), r_Height(kmax) )
    end do
    k = kmax
    r_DPHeight(k) = r_Height(kmax)

    ! search index
    !   Departure point is located in a r_DPIndex'th layer.
    !   Here, r_DPIndex is layer index, which ranges from 1 to kmax.
    do k = 0, kmax-1
      do kk = k+1, kmax
        if ( r_DPHeight(k) <= r_Height(kk) ) exit
      end do
      r_DPIndex(k) = kk
    end do
    k = kmax
    r_DPIndex(k) = -1

    ! p_{k} = p_{k-1} * exp( - dz / H )
    ! ln(p_{k}/p_{k-1}) = - dz / H
    ! H = - dz / ln(p_{k}/p_{k-1})
    !   = - (z_{k}-z_{k-1}) / ln(p_{k}/p_{k-1})
    do k = 0, kmax-1
      kk = r_DPIndex(k)
      !
      if ( r_Press(kk) <= 0.0d0 ) then
        PressTop = r_Press(kk-1) * 1.0d-1
      else
        PressTop = r_Press(kk  )
      end if
      PressBot  = r_Press (kk-1)
      HeightTop = r_Height(kk  )
      HeightBot = r_Height(kk-1)
      DPHeight  = r_DPHeight(k)
      !
      ScaleHeight = - ( HeightTop - HeightBot ) / log( PressTop / PressBot )
      r_DPPress(k) = PressBot * exp( - ( DPHeight - HeightBot ) / ScaleHeight )
    end do
    k = kmax
    r_DPPress(k) = r_Press(kmax)
    !
    do k = 1, kmax
      z_DelMassA(k) = 0.0d0
      !
      if ( k < kmax ) then
        kk = r_DPIndex(k  )
        z_DelMassA(k) = z_DelMassA(k) &
          & + z_DelMass(kk) &
          &   * ( r_Press(kk-1) - r_DPPress(k) ) / ( r_Press(kk-1) - r_Press(kk) )
      end if
      !
      do kk = r_DPIndex(k  )-1, r_DPIndex(k-1)+1, -1
        z_DelMassA(k) = z_DelMassA(k) + z_DelMass(kk)
      end do
      !
      kk = r_DPIndex(k-1)
      z_DelMassA(k) = z_DelMassA(k) &
        & + z_DelMass(kk) &
        &   * ( r_DPPress(k-1) - r_Press(kk) ) / ( r_Press(kk-1) - r_Press(kk) )
    end do


  end subroutine CloudMajCompSedimentation

  !----------------------------------------------------------------------------

  function a_SedVel( &
    & Grav, kmax, a_PartDen, a_PartDia, MolVisCoef, a_MeanFreePath &
    & )

    real(dp), intent(in ) :: Grav
    integer , intent(in ) :: kmax
    real(dp), intent(in ) :: a_PartDen(1:kmax)
    real(dp), intent(in ) :: a_PartDia(1:kmax)
    real(dp), intent(in ) :: MolVisCoef
    real(dp), intent(in ) :: a_MeanFreePath(1:kmax)

    real(dp) :: a_SedVel(1:kmax)

    a_SedVel =                                                      &
      & - a_PartDen * Grav * a_PartDia / ( 18.0_DP * MolVisCoef ) &
      &   * ( a_PartDia + 2.0_DP * a_MeanFreePath )

  end function a_SedVel

  !----------------------------------------------------------------------------

end module cloud_majcomp
