module refindex

  use vtype_module
  use ni3_module

  implicit NONE

  private

  public :: RefIndexInit
  public :: RefIndexFinalize
  public :: RefIndexCalc

  integer, save        :: NRITbl
  real(8), allocatable :: w_RITblWl  (:)
  real(8), allocatable :: w_RITblReal(:)
  real(8), allocatable :: w_RITblImag(:)

  logical, save        :: FlagCheckAndStop
  logical, save        :: FlagOutOfWaveNumRangeSmall
  logical, save        :: FlagOutOfWaveNumRangeLarge


contains

  !----------------------------------------------------------------------------

  subroutine RefIndexInit( &
    & NcFN, &
    & FlagStop &
    & )

    character(*), intent(in) :: NcFN
    logical     , intent(in) :: FlagStop


    ! Local variables
    !
    character(stdstr) :: Mode
    integer           :: NcID


    FlagCheckAndStop = FlagStop


    write( 6, * ) 'Reading ', trim( NcFN )

    Mode = "read"
    call ni3_open( NcFN, Mode, NcID )
    call ni3_inq_dimlen( NcID, "wl", NRITbl )
    allocate( w_RITblWl  (NRITbl) )
    allocate( w_RITblReal(NRITbl) )
    allocate( w_RITblImag(NRITbl) )
    call ni3_get_var( NcID, "wl", w_RITblWl )
    call ni3_get_var( NcID, "refindexreal", w_RITblReal )
    call ni3_get_var( NcID, "refindeximag", w_RITblImag )
    call ni3_close( NcID )

    if ( w_RITblWl(1) > w_RITblWl(2) ) then
      write( 6, * ) 'Order of refractive index is inappropriate: ', &
        & w_RITblWl(1), w_RITblWl(2)
      stop
    end if

    FlagOutOfWaveNumRangeSmall = .false.
    FlagOutOfWaveNumRangeLarge = .false.


  end subroutine RefIndexInit

  !----------------------------------------------------------------------------

  subroutine RefIndexFinalize


    FlagOutOfWaveNumRangeSmall = .false.
    FlagOutOfWaveNumRangeLarge = .false.

    deallocate( w_RITblWl   )
    deallocate( w_RITblReal )
    deallocate( w_RITblImag )


  end subroutine RefIndexFinalize

  !----------------------------------------------------------------------------

  subroutine RefIndexCalc( &
    & Wl,                  & ! (in)
    & RIReal, RIImag,      & ! (out)
    & Index                & ! (inout) optional
    & )

    real(8), intent(in   ) :: Wl
    real(8), intent(out  ) :: RIReal
    real(8), intent(out  ) :: RIImag
    integer, intent(inout), optional :: Index


    logical :: FlagWithinRange
    integer :: l
    integer :: IndexS
    real(8) :: RIImag1
    real(8) :: RIImag2


    FlagWithinRange = .true.

    if ( Wl < w_RITblWl(1) ) then
      if ( FlagOutOfWaveNumRangeSmall ) then
      else
        write( 6, * ) '  Specified wavelength is out of range.'
        write( 6, * ) '    Wavelength has to be larger than ', &
          & w_RITblWl(1)
        FlagOutOfWaveNumRangeSmall = .true.
      end if
      FlagWithinRange = .false.
    end if
    if ( w_RITblWl(NRITbl) < Wl ) then
      if ( FlagOutOfWaveNumRangeLarge ) then
      else
        write( 6, * ) '  Specified wavelength is out of range.'
        write( 6, * ) '    Wavelength has to be less than ', &
          & w_RITblWl(NRITbl)
        FlagOutOfWaveNumRangeLarge = .true.
      end if
      FlagWithinRange = .false.
    end if


    if ( FlagWithinRange ) then

      if ( present( Index ) ) then
        IndexS = Index
        l = max( IndexS, 2 )
        if ( Wl < w_RITblWl(l) ) then
          do l = IndexS, 2, -1
            if ( w_RITblWl(l) <= Wl ) exit
          end do
        end if
        IndexS = max( l, 2 )
      else
        IndexS = 2
      end if

      do l = IndexS, NRITbl
        if ( Wl <= w_RITblWl(l) ) exit
      end do

    ! see http://www.atmos.washington.edu/ice_optical_constants/
    ! for selection of interpolation method
!!$    RIReal =                                    &
!!$      &   ( w_RITblReal(l) - w_RITblReal(l-1) ) &
!!$      & / ( w_RITblWl  (l) - w_RITblWl  (l-1) ) &
!!$      & * ( Wl             - w_RITblWl  (l-1) ) &
!!$      & + w_RITblReal(l-1)
!!$    RIImag=                                     &
!!$      &   ( w_RITblImag(l) - w_RITblImag(l-1) ) &
!!$      & / ( w_RITblWl  (l) - w_RITblWl  (l-1) ) &
!!$      & * ( Wl             - w_RITblWl  (l-1) ) &
!!$      & + w_RITblImag(l-1)
      RIReal =                                       &
        &      ( w_RITblReal(l) - w_RITblReal(l-1) ) &
        & / log( w_RITblWl  (l) / w_RITblWl  (l-1) ) &
        & * log( Wl             / w_RITblWl  (l-1) ) &
        & + w_RITblReal(l-1)
!!$      RIImag =                                       &
!!$        &   log( w_RITblImag(l) / w_RITblImag(l-1) ) &
!!$        & / log( w_RITblWl  (l) / w_RITblWl  (l-1) ) &
!!$        & * log( Wl             / w_RITblWl  (l-1) ) &
!!$        & + log( w_RITblImag(l-1) )
      RIImag1 = max( w_RITblImag(l-1), 1.0d-30 )
      RIImag2 = max( w_RITblImag(l  ), 1.0d-30 )
      RIImag =                                       &
        &   log( RIImag2        / RIImag1          ) &
        & / log( w_RITblWl  (l) / w_RITblWl  (l-1) ) &
        & * log( Wl             / w_RITblWl  (l-1) ) &
        & + log( RIImag1 )
      RIImag = exp( RIImag )

      if ( present( Index ) ) then
        Index = l
      end if

    else
      if ( FlagCheckAndStop ) then
        stop
      else
        if ( Wl < w_RITblWl(1) ) then
          l = 1
        else if ( w_RITblWl(NRITbl) < Wl ) then
          l = NRITbl
        else
          write( 6, * ) 'In RefIndexCalc: Unexpected else'
          stop
        end if
        RIReal = w_RITblReal(l)
        RIImag = w_RITblImag(l)
      end if
    end if


  end subroutine RefIndexCalc

  !----------------------------------------------------------------------------

end module refindex
